# -*- Mode: Perl -*-

#
# Copyright (c) 2001, Bryan Jurish.  All rights reserved.
#
# This package is free software.  You may redistribute it
# and/or modify it under the same terms as Perl itself.
#

###############################################################
#
#      File: Hashed.pm
#    Author: Bryan Jurish <jurish@ling.uni-potsdam.de>
#
# Description: Quick set implementation
#
###############################################################
package Set::Hashed;
# System modules
#use Carp;
require Exporter;
# 3rd party exstensions
# user extension modules
@ISA       = qw(Exporter);
@EXPORT    = qw();
@EXPORT_OK = qw();

###############################################################
# Package-global variables
###############################################################
$VERSION = 0.07;

###############################################################
# Subroutines : Constructors / Destructors
###############################################################

#--------------------------------------------------------------
# new(@elts) -- new, intialize to @elts
#--------------------------------------------------------------
sub new {
  my $proto = shift;
  my $class = ref($proto) || $proto;
  my $self = { };
  bless $self, $class;
  $self->insert(@_) if (@_);
  return $self;
}

###############################################################
# Subroutines : Access
###############################################################

#--------------------------------------------------------------
# members()
#  + returns a list of elements
#--------------------------------------------------------------
sub members { return values(%{$_[0]}); }

#--------------------------------------------------------------
# member($elt), contains(@elts)
#  + returns a true value if $elt is in set, false otherwise
#--------------------------------------------------------------
sub member {
  #my ($self, $elt) = @_;
  return (defined($_[1]) && exists($_[0]->{$_[1]}));
}
sub contains {
  my $self = shift;
  return 0 unless (@_);
  grep {
    return 0 unless (defined($_) && exists($self->{$_}));
  } @_;
  return 1;
}

#--------------------------------------------------------------
# size()
#  + returns the number of elements in set
#--------------------------------------------------------------
sub size { return scalar(values(%{$_[0]})); }


###############################################################
# Subroutines : Single-Set Manipulation
###############################################################

#--------------------------------------------------------------
# insert(@elts), add(@elts)
#  + adds @elts to set
#--------------------------------------------------------------
sub insert {
  my $self = shift;
  my @values = grep { defined($_) } @_;  # be safe
  @$self{@values} = @values;
  return $self;
}
*add = \&insert;

#--------------------------------------------------------------
# remove(@elts), delete(@elts)
#  + delete @elts from set
#--------------------------------------------------------------
sub remove {
  my $self = shift;
  delete @$self{grep { defined($_) } @_};
  return $self;
}
*delete = \&remove;

#--------------------------------------------------------------
# clear()
#  + removes all elements
#--------------------------------------------------------------
sub clear {
  my $self = shift;
  %$self = ();
  return $self;
}

###############################################################
# Subroutines : Unary Set Operations
###############################################################

#--------------------------------------------------------------
# clone() => new Set::Hashed with identical membership
#--------------------------------------------------------------
sub clone {
  my $self = shift;
  return bless { %$self }, ref($self);
}

###############################################################
# Subroutines : Binary Set Operations
###############################################################

#--------------------------------------------------------------
# $s1->assign($s2) => makes $s1 membership-identical to $s2
#--------------------------------------------------------------
sub assign {
  %{$_[0]} = %{$_[1]};
  return $_[0];
}

#--------------------------------------------------------------
# union($s2) => new Set::Hashed representing union
#--------------------------------------------------------------
sub union {
  my ($self, $s2) = @_;
  return bless { %$self, %$s2 }, ref($self);
}

#--------------------------------------------------------------
# _union($s2) => destructive union
#--------------------------------------------------------------
sub _union {
  my ($self, $s2) = @_;
  @$self{values(%$s2)} = values(%$s2);
  return $self;
}


#--------------------------------------------------------------
# intersection($s2) => new Set::Hashed representing intersection
#--------------------------------------------------------------
sub intersection {
  my ($self,$s2) = @_;
  return $self->clone->_intersection($s2);
}
#sub intersection {
#  my ($self,$s2) = @_;
#  return $self->clone->_difference($self->difference($s2));
#}


#--------------------------------------------------------------
# _intersection($s2) => destructive intersection
#--------------------------------------------------------------
sub _intersection {
  my ($self,$s2) = @_;
  foreach (values(%$self)) {
    delete($self->{$_}) unless (exists($s2->{$_}));
  }
  return $self;
}
#sub _intersection {
#  my ($self,$s2) = @_;
#  return $self->_difference($self->difference($s2));
#}


#--------------------------------------------------------------
# difference($s2) => new Set::Hashed representing difference
#--------------------------------------------------------------
sub difference {
  my ($self,$s2) = @_;
  my $d = $self->clone;
  delete @$d{values(%$s2)};
  return $d;
}

#--------------------------------------------------------------
# _difference($s2) => destructive difference
#--------------------------------------------------------------
sub _difference {
  my ($self,$s2) = @_;
  delete @$self{values(%$s2)};
  return $self;
}


###############################################################
# Subroutines : Unary Set Comparison/Testing
###############################################################

#--------------------------------------------------------------
# is_null() => 1 if no members, 0 otherwise
#--------------------------------------------------------------
sub is_null { return (! %{$_[0]}); }
*is_empty = \&is_null;

###############################################################
# Subroutines : Binary Set Comparison
###############################################################

#--------------------------------------------------------------
# is_equal($s2) => 1 if membership is identical, 0 otherwise
#--------------------------------------------------------------
sub is_equal {
  my ($self,$s2) = @_;
  my ($e);
  foreach $e ($self->members) { return 0 unless ($s2->member($e)); }
  foreach $e ($s2->members) { return 0 unless ($self->member($e)); }
  return 1;
}

#--------------------------------------------------------------
# $self->is_subset($s2) => 1 if $self is a subset of $s2, 
#                          0 otherwise
#--------------------------------------------------------------
sub is_subset {
  my ($self,$s2) = @_;
  my ($e);
  foreach $e ($self->members) { return 0 unless ($s2->member($e)); }
  return 1;
}

#--------------------------------------------------------------
# $self->is_proper_subset($s2) => 1 if $self is a proper subset
#                                 of $s2, 0 otherwise
#--------------------------------------------------------------
sub is_proper_subset {
  my ($self,$s2) = @_;
  return 0 unless ($self->is_subset($s2));
  my ($e);
  foreach $e ($s2->members) { return 1 unless ($self->member($e)); }
  return 0;
}

# Autoload methods go after =cut, and are processed by the autosplit program.

1;
__END__
# Below is the stub of documentation for your module. You better edit it!

=head1 NAME

Set::Hashed - yet another Perl extension for set operations.

=head1 SYNOPSIS

  use Set::Hashed;

  $s = Set::Hashed->new();          # make a new set
  $s = Set::Hashed->new($e1, ...);  # ... with initial elements

  $s->insert($e1, ...);             # add elements to the set
  $s->remove($e1, ...);             # ... or remove them

  @elements = $s->members();        # list the elements
  $bool = $s->member($e1);          # ... or check membership
  $bool = $s->contains(@elts);      # ... for multiple elements
  $int = $s->size();                # ... or just see how many there are

  $t = $s->clone();                 # new set, duplicate membership
  $t->assign($s);                   # ... for existing sets $t

  $u = $s->union($t);               # $u is a new Set::Hashed
  $u = $s->_union($t);              # ... destructive union ($u eq $s)
  $i = $s->intersection($t);        # $i is a new Set::Hashed
  $i = $s->_intersection($t);       # ... destructive intersection ($i eq $s)
  $d = $s->difference($t);          # $d is a new Set::Hashed
  $d = $s->_difference($t);         # ... destructive difference ($d eq $s)

  $bool = $s->is_null();            # check if any members are present
  $bool = $s->is_equal($t);         # true if $s and $t have identical membership (no recursion)
  $bool = $s->is_subset($t);        # true if $s is a subset of $t (no recursion)
  $bool = $s->is_proper_subset($t); # true if $s is a proper subset of $t (no recursion)

=head1 DESCRIPTION

Set::Hashed provides set operations for scalars (including references) based
on an internal representation as a hash.  It provides a subset (sic) of the
functionality provided by the Set::Scalar module, and seems to be slightly
faster where the functionality overlaps.

=head1 BUGS

C<Set::Hashed> objects cannot contain C<undef> values -- C<undef> is
silently ignored by methods such as C<insert()>, and C<remove()>,
and caught by C<member()> and C<contains()>.

=head1 ACKNOLWDGEMENTS

perl by Larry Wall.

Much code adapted from the Set::Scalar module by Jarkko Hietaniemi E<lt>jhi@iki.fiE<gt>.

=head1 AUTHOR

Bryan Jurish E<lt>jurish@ling.uni-potsdam.deE<gt>

=head1 COPYRIGHT

Copyright (c) 2001, Bryan Jurish.  All rights reserved.

This package is free software.  You may redistribute it
and/or modify it under the same terms as Perl itself.

=head1 SEE ALSO

perl(1).
Set::Scalar(3pm).

=cut
